import random
from datetime import datetime, timedelta, date
from typing import List
from uuid import uuid4

import pytest
from nothion import NotionClient, PersonalStats
from tickthon import Task

from nothion._notion_parsers import NotionParsers
from nothion._notion_payloads import NotionPayloads
from nothion._notion_table_headers import ExpensesHeaders, StatsHeaders
from nothion.data_models.expense_log import ExpenseLog
from .conftest import NT_EXPENSES_DB_ID, NT_STATS_DB_ID, NT_TASKS_DB_ID, NT_NOTES_DB_ID


@pytest.fixture(scope="module")
def notion_client(notion_info):
    return NotionClient(notion_info["auth_secret"],
                        tasks_db_id=NT_TASKS_DB_ID,
                        stats_db_id=NT_STATS_DB_ID,
                        expenses_db_id=NT_EXPENSES_DB_ID,
                        notes_db_id=NT_NOTES_DB_ID)


def test_get_active_tasks(notion_client):
    active_tasks = notion_client.tasks.get_active_tasks()

    assert len(active_tasks) > 0
    assert isinstance(active_tasks, List) and all(isinstance(i, Task) for i in active_tasks)


def test_get_task(notion_client):
    expected_task = Task(ticktick_id="",
                         ticktick_etag="",
                         created_date="2025-07-14",
                         status=0,
                         title="Test Existing Task Static",
                         focus_time=0.9,
                         deleted=0,
                         tags=tuple(["test-existing-tag"]),
                         project_id="t542b6d8e9f2de3c5d6e7f8a9s2h",
                         timezone="",
                         due_date="9999-09-09",
                         column_id="4ff69f89b28d81f38d47",
                         )

    task = notion_client.tasks.get_task(expected_task)

    assert task == expected_task


def test_get_task_that_does_not_exist(notion_client):
    search_task = Task(ticktick_id="0testdoesntexisttask0",
                       ticktick_etag="0testdoesntexisttask0",
                       created_date="2099-09-09",
                       status=2,
                       title="Test Task That Does Not Exist",
                       )
    task = notion_client.tasks.get_task(search_task)

    assert task is None


def test_get_task_with_missing_properties(notion_client):
    expected_task = Task(ticktick_id="",
                         ticktick_etag="",
                         created_date="2025-07-15",
                         status=0,
                         title="Test Existing Task With Missing Data",
                         )

    task = notion_client.tasks.get_task(expected_task)

    assert task == expected_task


@pytest.mark.parametrize("task_title, expected_status", [
    # Test with a test task
    ("Test Existing Task Static", True),

    # Test with a task that does not exist
    ("0testtask0", False),
])
def test_is_task_already_created(notion_client, task_title, expected_status):
    is_task_created = notion_client.tasks.is_task_already_created(Task(ticktick_id="", ticktick_etag="", created_date="", title=task_title))

    assert is_task_created == expected_status


@pytest.mark.xfail(reason="The task creation in notion is currently disabled, because it's syncing automatically due to the integration with ticktick."
                          "I'll keep this tests here just in case I need to re-enable it in the future.")
def test_create_task(notion_client):
    task_id = uuid4().hex
    expected_task = Task(ticktick_id=task_id,
                         ticktick_etag="created-task-to-delete",
                         created_date="9999-09-09",
                         status=0,
                         title="Test Task to Delete",
                         focus_time=0.9,
                         tags=("test", "existing", "delete"),
                         project_id="a123a4b5c6d7e8f9a0b1c2d3s4h",
                         timezone="America/Bogota",
                         due_date="9999-09-09",
                         )

    notion_client.tasks.create(expected_task)

    task = notion_client.tasks.get_task(expected_task)
    assert task == expected_task

    notion_client.tasks.delete(expected_task)
    assert notion_client.tasks.is_already_created(expected_task) is False

@pytest.mark.xfail(reason="The task completion in notion is currently disabled, because it's syncing automatically due to the integration with ticktick."
                          "I'll keep this tests here just in case I need to re-enable it in the future.")
def test_complete_task(notion_client):
    task_id = uuid4().hex
    expected_task = Task(ticktick_id=task_id,
                         ticktick_etag="complete",
                         created_date="9999-09-09",
                         status=0,
                         title="Test Task to Complete",
                         focus_time=0.9,
                         tags=("test", "existing", "complete"),
                         project_id="f9ri34b5c6f7rh29a0b1f9eo2ln",
                         timezone="America/Bogota",
                         due_date="9999-09-09",
                         )

    notion_client.tasks.create(expected_task)
    notion_client.tasks.complete(expected_task)

    task = notion_client.tasks.get_task(expected_task)
    assert task.status == 2

    notion_client.tasks.delete(expected_task)
    assert notion_client.tasks.is_already_created(expected_task) is False


def test_update_task(notion_client):
    expected_task = Task(ticktick_id="",
                         ticktick_etag="",
                         created_date="2025-07-15",
                         status=0,
                         title="Test Existing Task to Update",
                         focus_time=random.random(),
                         tags=tuple(["test-existing-tag"]),
                         project_id="4a72b6d8e9f2103c5d6e7f8a9b0c",
                         column_id="4ff69f89b28d81f38d47",
                         due_date="9999-09-09",
                         )

    original_task = notion_client.tasks.get_task(expected_task)
    notion_client.tasks.update_task(expected_task)
    updated_task = notion_client.tasks.get_task(expected_task)

    assert updated_task == expected_task
    assert updated_task.title == original_task.title
    assert updated_task.focus_time != original_task.focus_time


def test_add_expense_log(notion_client):
    expected_expense_log = ExpenseLog(date="9999-09-09", expense=99.9, product="Test Expense Log")

    expense_log = notion_client.expenses.add_expense_log(expected_expense_log)

    expense_log_entry = notion_client.notion_api.get_table_entry(expense_log["id"])
    expense_log_properties = expense_log_entry["properties"]
    assert expense_log_properties[ExpensesHeaders.DATE.value]["date"]["start"] == expected_expense_log.date
    assert expense_log_properties[ExpensesHeaders.EXPENSE.value]["number"] == expected_expense_log.expense
    assert (expense_log_properties[ExpensesHeaders.PRODUCT.value]["title"][0]["text"]["content"]
            == expected_expense_log.product)

    notion_client.notion_api.update_table_entry(expense_log["id"], NotionPayloads.delete_table_entry())


def test_get_incomplete_stats_dates(notion_client):
    stats_date = datetime.now() + timedelta(days=2)

    incomplete_dates = notion_client.stats.get_incomplete_dates(stats_date)

    assert len(incomplete_dates) >= 2
    assert (isinstance(incomplete_dates, List) and
            all(datetime.strptime(i, '%Y-%m-%d') for i in incomplete_dates))


def test_create_stats_row(notion_client):
    stats = PersonalStats(date="9999-09-09", focus_total_time=1.0, focus_work_time=3.0, focus_personal_time=1.0, work_time=3.0,
                          leisure_time=4.0, sleep_time_amount=5.0, fall_asleep_time=6.0, sleep_score=7.0,
                          weight=8.0, steps=9.0, water_cups=10)

    notion_client.stats.update(stats, overwrite_stats=True)

    date_row = notion_client.notion_api.query_table(notion_client.stats_db_id, NotionPayloads.get_date_rows("9999-09-09"))[0]
    date_row_properties = date_row["properties"]
    assert date_row_properties[StatsHeaders.DATE.value]["date"]["start"] == stats.date
    assert date_row_properties[StatsHeaders.FOCUS_TOTAL_TIME.value]["number"] == stats.focus_total_time
    assert date_row_properties[StatsHeaders.FOCUS_PERSONAL_TIME.value]["number"] == stats.focus_personal_time
    assert date_row_properties[StatsHeaders.FOCUS_WORK_TIME.value]["number"] == stats.focus_work_time
    assert date_row_properties[StatsHeaders.WORK_TIME.value]["number"] == stats.work_time
    assert date_row_properties[StatsHeaders.LEISURE_TIME.value]["number"] == stats.leisure_time
    assert date_row_properties[StatsHeaders.SLEEP_TIME_AMOUNT.value]["number"] == stats.sleep_time_amount
    assert date_row_properties[StatsHeaders.SLEEP_DEEP_AMOUNT.value]["number"] == stats.sleep_deep_amount
    assert date_row_properties[StatsHeaders.FALL_ASLEEP_TIME.value]["number"] == stats.fall_asleep_time
    assert date_row_properties[StatsHeaders.SLEEP_SCORE.value]["number"] == stats.sleep_score
    assert date_row_properties[StatsHeaders.WEIGHT.value]["number"] == stats.weight
    assert date_row_properties[StatsHeaders.STEPS.value]["number"] == stats.steps
    assert date_row_properties[StatsHeaders.WATER_CUPS.value]["number"] == stats.water_cups


    notion_client.notion_api.update_table_entry(date_row["id"], NotionPayloads.delete_table_entry())


def test_update_stats_row(notion_client):
    notion_api = notion_client.notion_api
    expected_stat = PersonalStats(date="1999-09-09",
                                  focus_total_time=random.random(),
                                  focus_personal_time=4.5,
                                  focus_work_time=5.5,
                                  work_time=1.2,
                                  fall_asleep_time=3,
                                  sleep_time_amount=4,
                                  sleep_score=89,
                                  leisure_time=3.4,
                                  weight=70.0,
                                  steps=1231,
                                  water_cups=8,
                                  )

    original_stat = NotionParsers.parse_stats_rows([notion_api.get_table_entry("c568738e82a24b258071e5412db89a2f")])[0]
    notion_client.stats.update(expected_stat, overwrite_stats=True)
    updated_stat = NotionParsers.parse_stats_rows([notion_api.get_table_entry("c568738e82a24b258071e5412db89a2f")])[0]

    assert updated_stat == expected_stat
    assert updated_stat.date == original_stat.date
    assert updated_stat.focus_total_time != original_stat.focus_total_time


@pytest.mark.parametrize("start_date, end_date, expected_stats", [
    # Test start date before end date
    (date(2023, 1, 1), date(2023, 1, 3),
     [PersonalStats(date='2023-01-01', work_time=2.03, leisure_time=6.5, focus_total_time=0, focus_work_time=1.97, focus_personal_time=1.23, weight=0, water_cups=0),
      PersonalStats(date='2023-01-02', work_time=3.24, leisure_time=3.24, focus_total_time=3.12, focus_work_time=3.12, focus_personal_time=0.45, weight=0, water_cups=0),
      PersonalStats(date='2023-01-03', work_time=7.57, leisure_time=1.51, focus_total_time=6.33, focus_work_time=7.42, focus_personal_time=2.07, weight=0, water_cups=0)]),

    # Test start date equal to end date
    (date(2023, 1, 1), date(2023, 1, 1),
     [PersonalStats(date='2023-01-01', work_time=2.03, leisure_time=6.5, focus_total_time=0, focus_work_time=1.97, focus_personal_time=1.23, weight=0, water_cups=0)]),

    # Test start date after end date
    (date(2023, 1, 3), date(2023, 1, 1), []),
])
def test_get_stats_between_dates(notion_client, start_date, end_date, expected_stats):
    stats = notion_client.stats.get_between_dates(start_date, end_date)
    assert stats == expected_stats


@pytest.mark.parametrize("title, page_type, expected_result", [
    # Test with a valid title and page type
    ("test-page-entry", "note", True),

    # Test with a valid title and a wrong page type
    ("test-journal-entry", "test-wrong-page-type", False),

    # Test with a wrong title and a valid page type
    ("test-wrong-title", "journal", False),

    # Test with a wrong title and page type
    ("test-wrong-title", "test-wrong-page-type", False),
])
def test_is_note_page_already_created(notion_client, title, page_type, expected_result):
    is_note_page_created = notion_client.notes.is_page_already_created(title, page_type)
    assert is_note_page_created == expected_result



from enum import Enum


class ExpensesHeaders(Enum):
    PRODUCT = "item"
    EXPENSE = "expense"
    DATE = "date"


class TasksHeaders(Enum):
    DONE = "Checkbox"
    TITLE = "Title"
    FOCUS_TIME = "Focus time"
    DATE = "Date"
    CREATED_DATE = "Created time"
    TAGS = "Tag"
    PROJECT_ID = "List id"
    COLUMN_ID = "Column id"


class StatsHeaders(Enum):
    COMPLETED = "completed"
    DATE = "date"
    FOCUS_TOTAL_TIME = "ftt - focus time total"
    FOCUS_PERSONAL_TIME = "ftp - focus time personal"
    FOCUS_WORK_TIME = "ftw - focus time work"
    WORK_TIME = "ftr - focus time rescuetime"
    LEISURE_TIME = "lt - leisure time"
    SLEEP_TIME_AMOUNT = "sa - sleep amount"
    SLEEP_DEEP_AMOUNT = "sda - sleep deep amount"
    FALL_ASLEEP_TIME = "st - fall asleep time"
    SLEEP_SCORE = "ss - sleep score"
    WEIGHT = "kg - weight"
    STEPS = "stp - steps"
    WATER_CUPS = "wc - water cups"


class NotesHeaders(Enum):
    NOTE = "Note"
    TYPE = "Type"
    SUBTYPE = "Sub-type"
    DUE_DATE = "Due date"



import json
import random
from typing import List, Dict

import pytest

from nothion._notion_api import NotionAPI
from nothion._notion_payloads import NotionPayloads
from .conftest import (NT_TASKS_DB_ID, TEST_TASK, TEST_STAT, TEST_EXPENSE_LOG, EXISTING_TEST_TASK_PAGE_ID,
                       NT_NOTES_DB_ID, NT_STATS_DB_ID, NT_EXPENSES_DB_ID, EXISTING_TEST_STAT_PAGE_ID,
                       EXISTING_TEST_EXPENSE_LOG_PAGE_ID, EXISTING_TEST_JOURNAL_PAGE_ID)


notion_payloads = NotionPayloads(tasks_db_id=NT_TASKS_DB_ID, 
                                 notes_db_id=NT_NOTES_DB_ID, 
                                 stats_db_id=NT_STATS_DB_ID, 
                                 expenses_db_id=NT_EXPENSES_DB_ID)

@pytest.fixture(scope="module")
def notion_api(notion_info):
    return NotionAPI(notion_info["auth_secret"])


@pytest.mark.parametrize("payload", [
    # Test with tasks database
    (notion_payloads.create_task(TEST_TASK)),

    # Test with stats database
    (notion_payloads.update_stats_row(TEST_STAT, new_row=True)),

    # Test with expenses database
    (notion_payloads.create_expense_log(TEST_EXPENSE_LOG)),
])
def test_create_table_entry(notion_api, payload):
    table_entry_data = notion_api.create_table_entry(payload)

    table_entry = notion_api.get_table_entry(table_entry_data["id"])

    notion_api.update_table_entry(table_entry["id"], notion_payloads.delete_table_entry())


@pytest.mark.parametrize("database_id, query, expected_property, expected_value", [
    # Test with tasks database
    (NT_TASKS_DB_ID,
     {"filter":{"and":[{"property":"Title","title":{"equals":"Test Existing Task Static"}}]}},
     "id", "23041ab9-8366-810b-9cfe-dae0f672e9b1"),

    # Test with tasks database
    (NT_NOTES_DB_ID,
     {"filter": {"and": [{"property": "Note", "title": {"equals": "Test Existing Task"}}]}},
     "id", "c964714a-6fd8-474a-ba60-bb215c5ce77b"),

    # Test with stats database
    (NT_STATS_DB_ID,
     {"filter": {"and": [{"property": "name", "rich_text": {"equals": "Test Existing Stat"}}]}},
     "id", "c568738e-82a2-4b25-8071-e5412db89a2f"),

    # Test with expenses database
    (NT_EXPENSES_DB_ID,
     {"filter": {"and": [{"property": "item", "rich_text": {"equals": "Test Existing Expense Log"}}]}},
     "id", "36de61f8-b24c-49e2-86bb-5b0aca9740ab"),
])
def test_query_table(notion_api, database_id, query, expected_property, expected_value):
    table_entries = notion_api.query_table(database_id, query)

    assert len(table_entries) == 1
    assert table_entries[0][expected_property] == expected_value


@pytest.mark.parametrize("query, expected_pages", [
    # Test with a page limit
    ({"filter": {"and": [{"property": "año", "formula": {"number": {"equals": 2023}}},
                         {"property": "mes", "formula": {"number": {"equals": 1}}}]},
      "page_size": 10},
     10),

    # Test without a page limit
    ({"filter": {"and": [{"property": "año", "formula": {"number": {"equals": 2023}}}]}},
     361),
])
def test_query_with_multiple_pages(notion_api, query, expected_pages):
    table_entries = notion_api.query_table(NT_EXPENSES_DB_ID, query)

    assert isinstance(table_entries, List) and all(isinstance(i, Dict) for i in table_entries)
    assert len(table_entries) == expected_pages


@pytest.mark.parametrize("page_id, stable_property, updated_property, payload", [
    # Test with tasks database
    (EXISTING_TEST_TASK_PAGE_ID, "Note", "Focus time",
     json.dumps({"properties": {"Focus time": {"number": random.random()}}})),

    # Test with stats database
    (EXISTING_TEST_STAT_PAGE_ID, "ftr - focus time rescuetime", "ftw - focus time work",
     json.dumps({"properties": {"ftw - focus time work": {"number": random.random()}}})),

    # Test with expenses database
    (EXISTING_TEST_EXPENSE_LOG_PAGE_ID, "item", "expense",
     json.dumps({"properties": {"expense": {"number": random.random()}}})),
])
def test_update_table_entry(notion_api, stable_property, updated_property, page_id, payload):
    original_table_entry = notion_api.get_table_entry(page_id)

    notion_api.update_table_entry(page_id, payload)

    updated_table_entry = notion_api.get_table_entry(page_id)
    assert original_table_entry["properties"][stable_property] == updated_table_entry["properties"][stable_property]
    assert original_table_entry["properties"][updated_property] != updated_table_entry["properties"][updated_property]


def test_get_block_children(notion_api):
    expected_children = 6
    expected_first_children_type = "heading_1"
    block_children = notion_api.get_block_children(EXISTING_TEST_JOURNAL_PAGE_ID)

    assert len(block_children.get("results", [])) == expected_children
    assert block_children.get("results", [])[0].get("type", "") == expected_first_children_type



from datetime import datetime, timedelta
from typing import List, Optional

from tickthon import Task

from nothion import PersonalStats
from nothion._notion_parsers import NotionParsers
from nothion._notion_payloads import NotionPayloads
from nothion._notion_api import NotionAPI
from nothion.data_models.expense_log import ExpenseLog


class NotionClient:

    def __init__(self,
                 auth_secret: str,
                 tasks_db_id: str = "",
                 stats_db_id: str = "",
                 notes_db_id: str = "",
                 expenses_db_id: str = ""):
        self.notion_api = NotionAPI(auth_secret)
        self.active_tasks: List[Task] = []

        self.tasks_db_id = tasks_db_id
        self.stats_db_id = stats_db_id
        self.notes_db_id = notes_db_id
        self.expenses_db_id = expenses_db_id

        self.notion_payloads = NotionPayloads(
            tasks_db_id=self.tasks_db_id,
            stats_db_id=self.stats_db_id,
            notes_db_id=self.notes_db_id,
            expenses_db_id=self.expenses_db_id
        )

        self.tasks = self.TasksHandler(self)
        self.notes = self.NotesHandler(self)
        self.stats = self.StatsHandler(self)
        self.expenses = self.ExpensesHandler(self)
        self.blocks = self.BlocksHandler(self)

    class TasksHandler:
        def __init__(self, client):
            self.client = client

        def get_active_tasks(self) -> List[Task]:
            """Gets all active tasks from Notion that are not done."""
            payload = self.client.notion_payloads.get_active_tasks()
            raw_tasks = self.client.notion_api.query_table(self.client.tasks_db_id, payload)
            notion_tasks = NotionParsers.parse_notion_tasks(raw_tasks)
            self.client.active_tasks = notion_tasks
            return notion_tasks

        def get_task(self, ticktick_task: Task) -> Optional[Task]:
            """Gets the task from Notion that have the given ticktick etag."""
            payload = self.client.notion_payloads.get_notion_task(ticktick_task)
            raw_tasks = self.client.notion_api.query_table(self.client.tasks_db_id, payload)

            notion_tasks = NotionParsers.parse_notion_tasks(raw_tasks)
            if notion_tasks:
                return notion_tasks[0]
            return None

        def get_notion_id(self, ticktick_task: Task) -> str:
            """Gets the Notion ID of a task."""
            payload = self.client.notion_payloads.get_notion_task(ticktick_task)
            raw_tasks = self.client.notion_api.query_table(self.client.tasks_db_id, payload)

            return raw_tasks[0]["id"].replace("-", "")

        def is_task_already_created(self, task: Task) -> bool:
            """Checks if a task is already created in Notion."""
            payload = self.client.notion_payloads.get_notion_task(task)
            raw_tasks = self.client.notion_api.query_table(self.client.tasks_db_id, payload)
            return len(raw_tasks) > 0

        def create(self, task: Task) -> Optional[dict]:
            """Creates a task in Notion."""
            payload = self.client.notion_payloads.create_task(task)

            if not self.is_task_already_created(task):
                return self.client.notion_api.create_table_entry(payload)
            return None

        def update_task(self, task: Task):
            """Updates a task in Notion."""
            page_id = self.get_notion_id(task)
            payload = self.client.notion_payloads.update_task(task)
            self.client.notion_api.update_table_entry(page_id, payload)

        def complete(self, task: Task):
            """Completes a task in Notion."""
            page_id = self.get_notion_id(task)
            payload = self.client.notion_payloads.complete_task()
            self.client.notion_api.update_table_entry(page_id, payload)

        def delete(self, task: Task):
            """Deletes a task from Notion."""
            task_payload = self.client.notion_payloads.get_notion_task(task)
            raw_tasks = self.client.notion_api.query_table(self.client.tasks_db_id, task_payload)

            delete_payload = self.client.notion_payloads.delete_table_entry()
            for raw_task in raw_tasks:
                page_id = raw_task["id"]
                self.client.notion_api.update_table_entry(page_id, delete_payload)

    class NotesHandler:
        def __init__(self, client):
            self.client = client

        def is_page_already_created(self, title: str, page_type: str) -> bool:
            """Checks if a note's page is already created in Notion."""
            payload = self.client.notion_payloads.get_note_page(title, page_type)
            raw_tasks = self.client.notion_api.query_table(self.client.notes_db_id, payload)
            return len(raw_tasks) > 0

        def create_page(self,
                        title: str,
                        page_type: str,
                        page_subtype: tuple[str],
                        date: datetime,
                        content: str) -> dict | None:
            """Creates a note page in Notion."""
            payload = self.client.notion_payloads.create_note_page(title, page_type, page_subtype, date, content)

            if not self.is_page_already_created(title, page_type):
                return self.client.notion_api.create_table_entry(payload)
            return None

    class StatsHandler:
        def __init__(self, client):
            self.client = client
            self.payloads = self.client.notion_payloads

        def _get_last_row_checked(self) -> Optional[PersonalStats]:
            """Gets the last checked row from the stats in Notion database."""
            checked_rows = self.client.notion_api.query_table(self.client.stats_db_id,
                                                              self.payloads.get_checked_stats_rows())

            if checked_rows:
                return NotionParsers.parse_stats_rows([checked_rows[-1]])[0]
            return None

        def get_incomplete_dates(self, limit_date: datetime) -> List[str]:
            """Gets the dates that are incomplete in the stats database."""
            initial_date = datetime(limit_date.year, 1, 1)
            last_checked_row = self._get_last_row_checked()
            if last_checked_row:
                current_date = datetime.strptime(last_checked_row.date, "%Y-%m-%d")
                initial_date = current_date - timedelta(days=14)

            dates = []
            delta = limit_date - initial_date
            for delta_days in range(delta.days + 1):
                day = initial_date + timedelta(days=delta_days)
                dates.append(day.strftime("%Y-%m-%d"))

            return dates

        def update(self, stat_data: PersonalStats, overwrite_stats: bool = False):
            """Updates a row in the stats database in Notion."""
            raw_date_row = self.client.notion_api.query_table(self.client.stats_db_id,
                                                              self.payloads.get_date_rows(stat_data.date))

            date_row = NotionParsers.parse_stats_rows(raw_date_row)

            if date_row:
                row_id = raw_date_row[0]["id"]
                self.client.notion_api.update_table_entry(row_id,
                                                          self.payloads.update_stats_row(stat_data,
                                                                                         old_stats=date_row,
                                                                                         overwrite_stats=overwrite_stats
                                                                                         ))
            else:
                self.client.notion_api.create_table_entry(self.payloads.update_stats_row(stat_data,
                                                                                         new_row=True))

        def get_between_dates(self, start_date: datetime, end_date: datetime) -> List[PersonalStats]:
            """Gets stats between two dates."""
            raw_data = self.client.notion_api.query_table(self.client.stats_db_id,
                                                          self.payloads.get_data_between_dates(start_date,
                                                                                               end_date))
            return NotionParsers.parse_stats_rows(raw_data)

    class BlocksHandler:
        def __init__(self, client):
            self.client = client

        def get_all_children(self, block_id: str) -> list:
            """Recursively gets the children of a block in Notion."""
            children = []
            children_data = self.client.notion_api.get_block_children(block_id)

            for child in children_data.get("results", []):
                parsed_child: list[str | list] = [NotionParsers.parse_block(child)]
                children.append(parsed_child)

                if child.get("has_children", False):
                    parsed_child.append(self.get_all_children(child["id"]))

            return children

    class ExpensesHandler:
        def __init__(self, client):
            self.client = client

        def add_expense_log(self, expense_log: ExpenseLog) -> dict:
            """Adds an expense log to the expenses DB in Notion."""
            payload = self.client.notion_payloads.create_expense_log(expense_log)
            return self.client.notion_api.create_table_entry(payload)



from tickthon import Task

from nothion._notion_table_headers import StatsHeaders, TasksHeaders
from nothion.data_models.personal_stats import PersonalStats


class NotionParsers:

    @staticmethod
    def _get_rich_text(properties: dict, field: str) -> str:
        """Extracts plain text from rich text field."""
        try:
            rich_text = properties[field]["rich_text"]
            return rich_text[0]["plain_text"] if rich_text else ""
        except (KeyError, IndexError):
            return ""

    @staticmethod
    def _get_date(properties: dict, field: str) -> str:
        """Extracts start date from date field."""
        try:
            date_field = properties[field]

            if date_field.get("date"):
                parsed_date = date_field["date"]["start"]
            elif date_field.get("created_time"):
                parsed_date = date_field["created_time"].split("T")[0]
            elif date_field.get("last_edited_time"):
                parsed_date = date_field["last_edited_time"].split("T")[0]
            else:
                parsed_date = ""

            return parsed_date
        except KeyError:
            return ""

    @staticmethod
    def _get_title(properties: dict, field: str) -> str:
        """Extracts plain text from title field."""
        try:
            title = properties[field]["title"]
            return title[0]["plain_text"] if title else ""
        except (KeyError, IndexError):
            return ""

    @staticmethod
    def _get_number(properties: dict, field: str) -> float:
        """Extracts number from number field."""
        try:
            return properties[field]["number"] or 0
        except KeyError:
            return 0

    @classmethod
    def parse_notion_tasks(cls, raw_tasks: list[dict]) -> list[Task]:
        """Parses the raw tasks from Notion into Task objects.

        Args:
            raw_tasks: A single task dict or list of task dicts from Notion API

        Returns:
            List[Task]: List of parsed Task objects
        """

        parsed_tasks = []
        for raw_task in raw_tasks:
            try:
                task_properties = raw_task["properties"]

                parsed_tasks.append(Task(
                    title=cls._get_title(task_properties, TasksHeaders.TITLE.value),
                    status=2 if task_properties.get(TasksHeaders.DONE.value, {}).get("checkbox", False) else 0,
                    ticktick_id="",
                    ticktick_etag="",
                    column_id=cls._get_rich_text(task_properties, TasksHeaders.COLUMN_ID.value),
                    created_date=cls._get_date(task_properties, TasksHeaders.CREATED_DATE.value),
                    focus_time=cls._get_number(task_properties, TasksHeaders.FOCUS_TIME.value),
                    deleted=int(raw_task.get("archived", False)),
                    tags=tuple(tag["name"] for tag in
                               task_properties.get(TasksHeaders.TAGS.value, {}).get("multi_select", [])),
                    project_id=cls._get_rich_text(task_properties, TasksHeaders.PROJECT_ID.value),
                    due_date=cls._get_date(task_properties, TasksHeaders.DATE.value)
                ))
            except Exception as e:
                print(f"Error parsing task: {e}")
                continue

        return parsed_tasks

    @classmethod
    def parse_stats_rows(cls, rows: list[dict]) -> list[PersonalStats]:
        """Parses the raw stats rows from Notion into PersonalStats objects."""

        rows_parsed = []
        for row in rows:
            try:
                row_properties = row["properties"]

                rows_parsed.append(
                    PersonalStats(
                        date=cls._get_date(row_properties, StatsHeaders.DATE.value),
                        focus_total_time=cls._get_number(row_properties, StatsHeaders.FOCUS_TOTAL_TIME.value),
                        focus_personal_time=cls._get_number(row_properties, StatsHeaders.FOCUS_PERSONAL_TIME.value),
                        focus_work_time=cls._get_number(row_properties, StatsHeaders.FOCUS_WORK_TIME.value),
                        work_time=cls._get_number(row_properties, StatsHeaders.WORK_TIME.value),
                        leisure_time=cls._get_number(row_properties, StatsHeaders.LEISURE_TIME.value),
                        sleep_time_amount=cls._get_number(row_properties, StatsHeaders.SLEEP_TIME_AMOUNT.value),
                        sleep_deep_amount=cls._get_number(row_properties, StatsHeaders.SLEEP_DEEP_AMOUNT.value),
                        fall_asleep_time=cls._get_number(row_properties, StatsHeaders.FALL_ASLEEP_TIME.value),
                        sleep_score=cls._get_number(row_properties, StatsHeaders.SLEEP_SCORE.value),
                        weight=cls._get_number(row_properties, StatsHeaders.WEIGHT.value),
                        steps=cls._get_number(row_properties, StatsHeaders.STEPS.value),
                        water_cups=int(cls._get_number(row_properties, StatsHeaders.WATER_CUPS.value)),
                    )
                )
            except Exception as e:
                print(f"Error parsing stats row: {e}")
                continue

        return rows_parsed

    @classmethod
    def parse_block(cls, block: dict) -> str:
        """Parses a block from Notion into a string."""
        try:
            block_type = block["type"]
            block_text = cls._get_rich_text(block[block_type], "rich_text")

            if block_type in ["paragraph", "heading_1", "heading_2", "heading_3"]:
                return block_text
            elif block_type == "bulleted_list_item":
                return "- " + block_text
            elif block_type == "toggle":
                return "> " + block_text
            return block_type
        except Exception as e:
            print(f"Error parsing block: {e}")
            return ""



import json
from datetime import datetime
from typing import Optional

from tickthon import Task

from nothion._notion_table_headers import TasksHeaders, ExpensesHeaders, StatsHeaders, NotesHeaders
from nothion.data_models.expense_log import ExpenseLog
from nothion.data_models.personal_stats import PersonalStats


class NotionPayloads:
    def __init__(self, tasks_db_id: str, expenses_db_id: str, stats_db_id: str, notes_db_id: str):
        self.tasks_db_id = tasks_db_id
        self.expenses_db_id = expenses_db_id
        self.stats_db_id = stats_db_id
        self.notes_db_id = notes_db_id

    @staticmethod
    def get_active_tasks() -> dict:
        tag_values = ["work-project", "dwtask", "swtask", "personal-project", "dtask", "stask", "wørk-focus-meeting"]

        return {
            "filter": {
                "or": [
                    {
                        "property": "Tag",
                        "multi_select": {
                            "contains": tag
                        }
                    } for tag in tag_values
                ]
            }
        }

    @staticmethod
    def _base_task_payload(task: Task) -> dict:
        payload = {
            "properties": {
                TasksHeaders.DONE.value: {"checkbox": task.status != 0},
                "title": {"title": [{"text": {"content": task.title}}]},
                TasksHeaders.FOCUS_TIME.value: {"number": task.focus_time},
                TasksHeaders.TAGS.value: {"multi_select": list(map(lambda tag: {"name": tag}, task.tags))},
                TasksHeaders.COLUMN_ID.value: {"rich_text": [{"text": {"content": task.column_id}}]},
                TasksHeaders.PROJECT_ID.value: {"rich_text": [{"text": {"content": task.project_id}}]},
                TasksHeaders.CREATED_DATE.value: {"date": {"start": task.created_date}},
            }
        }

        if task.due_date:
            payload["properties"][TasksHeaders.DATE.value] = {"date": {"start": task.due_date}}

        return payload

    def create_task(self, task: Task) -> str:
        payload = self._base_task_payload(task)
        payload["parent"] = {"database_id": self.tasks_db_id}
        return json.dumps(payload)

    @classmethod
    def update_task(cls, task: Task) -> str:
        payload = {
            "properties": {
                TasksHeaders.FOCUS_TIME.value: {"number": task.focus_time},
                TasksHeaders.COLUMN_ID.value: {"rich_text": [{"text": {"content": task.column_id}}]},
                TasksHeaders.PROJECT_ID.value: {"rich_text": [{"text": {"content": task.project_id}}]},
            }
        }

        return json.dumps(payload)

    @classmethod
    def complete_task(cls) -> str:
        payload = {"properties": {TasksHeaders.DONE.value: {"checkbox": True}}}
        return json.dumps(payload)

    @staticmethod
    def delete_table_entry() -> str:
        payload = {"archived": True}

        return json.dumps(payload)

    @staticmethod
    def get_notion_task(task: Task) -> dict:
        """Payload to get a notion task by its ticktick id or etag.

        Args:
            task: The task to search for.
        """
        payload = {
            "sorts": [{"property": TasksHeaders.DATE.value, "direction": "ascending"}],
            "filter": {
                "and": [
                    {"property": TasksHeaders.TITLE.value, "rich_text": {"equals": task.title}}
                ]
            }
        }

        return payload

    def create_expense_log(self, expense_log: ExpenseLog) -> str:
        payload = {
            "parent": {"database_id": self.expenses_db_id},
            "properties": {
                ExpensesHeaders.PRODUCT.value: {"title": [{"text": {"content": expense_log.product}}]},
                ExpensesHeaders.EXPENSE.value: {"number": expense_log.expense},
                ExpensesHeaders.DATE.value: {"date": {"start": expense_log.date}}
            }
        }

        return json.dumps(payload)

    @staticmethod
    def get_checked_stats_rows() -> dict:
        payload = {
            "sorts": [{"property": StatsHeaders.DATE.value, "direction": "descending"}],
            "filter": {"and": [{"property": StatsHeaders.COMPLETED.value, "checkbox": {"equals": True}}]},
            "page_size": 1
        }
        return payload

    @staticmethod
    def get_data_between_dates(initial_date: Optional[datetime], today_date: datetime) -> dict:
        filters = []
        if initial_date:
            filters.append({"property": "date", "date": {"on_or_after": initial_date.strftime("%Y-%m-%d")}})

        filters.append({"property": "date", "date": {"on_or_before": today_date.strftime("%Y-%m-%d")}})

        return {"sorts": [{"property": "day #", "direction": "ascending"}], "filter": {"and": filters}}

    @staticmethod
    def get_date_rows(date: str) -> dict:
        return {"filter": {"and": [{"property": "date", "date": {"equals": date}}]}}

    def update_stats_row(self,
                         stat: PersonalStats,
                         old_stats: PersonalStats | None = None,
                         new_row: bool = False,
                         overwrite_stats: bool = False) -> str:
        payload: dict = {
            "properties": {
                StatsHeaders.DATE.value: {"date": {"start": stat.date}}
            }
        }

        stats_fields = [header for header in StatsHeaders if header != StatsHeaders.DATE]
        if new_row or overwrite_stats:

            for header in stats_fields:
                attr_name = header.name.lower()
                new_value = getattr(stat, attr_name, None)

                if new_value is not None:
                    payload["properties"][header.value] = {"number": new_value}

        elif old_stats:

            for header in stats_fields:
                attr_name = header.name.lower()

                old_value = getattr(old_stats, attr_name, None)
                new_value = getattr(stat, attr_name, None)

                if old_value and not new_value:
                    payload["properties"][header.value] = {"number": old_value}

        if new_row:
            payload["parent"] = {"database_id": self.stats_db_id}

        return json.dumps(payload)

    def create_note_page(self, title: str, page_type: str, page_subtype: tuple[str], date: datetime,
                         content: str) -> str:
        content_block = {"object": "block",
                         "type": "paragraph",
                         "paragraph": {"rich_text": [{"type": "text", "text": {"content": content}}]}
                         }

        payload = {
            "parent": {"database_id": self.notes_db_id},
            "properties": {
                NotesHeaders.NOTE.value: {"title": [{"text": {"content": title}}]},
                NotesHeaders.TYPE.value: {"select": {"name": page_type}},
                NotesHeaders.SUBTYPE.value: {"multi_select": [{"name": st} for st in page_subtype]},
                NotesHeaders.DUE_DATE.value: {"date": {"start": date.strftime("%Y-%m-%d")}},
            },
            "children": [content_block]
        }

        return json.dumps(payload)

    @classmethod
    def get_note_page(cls, title: str, page_type: str) -> dict:
        return {"filter": {
                    "and": [{"property": NotesHeaders.NOTE.value,
                             "rich_text": {"equals": title}},
                            {"property": NotesHeaders.TYPE.value,
                             "select": {"equals": page_type}
                             },
                            ]
                        }
                }



from attrs import define


@define
class PersonalStats:
    """Represents a personal stats row in Notion.

    Attributes:
        date: The date of the stats in format YYYY-MM-DD.
        all other attributes are self-explanatory.
    """
    date: str
    focus_total_time: float | None
    focus_personal_time: float | None
    focus_work_time: float | None
    work_time: float | None
    leisure_time: float | None
    sleep_time_amount: float | None = 0.0
    sleep_deep_amount: float | None = 0.0
    fall_asleep_time: float | None = 0.0
    sleep_score: float | None = 0.0
    weight: float | None = 0.0
    steps: float | None = 0.0
    water_cups: int | None = None
